
#include "DataPage.hpp"

/** Constructor */
DataPage::DataPage():
  pagesize(DEFAULT_PAGE_SIZE),npage(DEFAULT_NPAGE),
  ipage_writing(-1)
{
  mutex.lock();  
  //  dat.resize(npage);
  tarr.resize(npage);
  status.resize(npage);
  for(int i=0; i<npage; i++){
    //    dat[i] = new uint8_t[pagesize];
    dat[i].resize(pagesize);
  }
  mutex.unlock();
  InitPages();
}


/** Constructor */
DataPage::DataPage(const int npg):
  pagesize(DEFAULT_PAGE_SIZE),npage(npg),
  ipage_writing(-1)
{
  
  //  dat.resize(npage);
  mutex.lock();
  tarr.resize(npage);
  status.resize(npage);
  for(int i=0; i<npage; i++){
    //    dat[i] = new uint8_t[pagesize];
    dat[i].resize(pagesize);
  }
  mutex.unlock();
  InitPages();
}


/** Destructor */
DataPage::~DataPage(){
//   for(int i=0; i < (int)dat.size(); i++){
//     if(dat[i] != 0) delete [] dat[i];
//     dat[i] =0;
//   }
}
const int DataPage::PageResize(const uint32_t pgsz ){
  mutex.lock();
  if(pagesize != pgsz){
    pagesize = pgsz;  
    
    for(int i=0; i < npage; i++){
      //    delete [] dat[i];
      //    dat[i] = new uint8_t[pagesize];
      dat[i].resize(pagesize);
    }
  }
  mutex.unlock();
  InitPages();  
  return 0;
}

// void DataPage::Set_Npage(const int n){
//   if(n == (int)npage) return;
//   for(uint i=0; i<npage; i++){    delete [] dat[i];  }
//   dat.resize(n); tarr.resize(n);   status.resize(npage);
//   npage = n;
//   for(uint i=0; i<npage; i++){    
//     dat[i] = new uint8_t[pagesize];
//   }
//   fill(status.begin(),status.end(),EMPTY);
// }

/** Returns Earliest epoch available from the page.
 * @return  
 *   - 0: If data is not ready for all pages.
 *   - time_t: If any of the page are ready.
 */

time_t DataPage::EarliestEpoch(){
  time_t t_tmp(0);
  bool flag (true);
  //  int ipage;
  for(int i=0; i < (int)tarr.size(); i++){
    if( (status[i] == WR_OK || status[i] == RD_OK)){
      if(flag){ t_tmp = tarr[i]; flag=false;}
      if(t_tmp > tarr[i]) t_tmp = tarr[i];
    }
  }
  if(flag){
    Logsystem::debug_logger()->Tstamp() << "At DataPage::EarliestEpoch(): failed, " 
					<< (int)tarr.size() << StrPageState() << "\n";
  }
  return t_tmp;
}

/** WStepPage()
 * Change the page state to write the data into the page. 
 * input: 
 *   -  state_dat: State value to be written to current page
 *   -  wmode:     Waiting mode, Wait until the data is read at leaset one time, or
 *                 NonWait: data on the page is overwritten regardless whether 
 *                 the data is read or not.
 */
int DataPage::WStepPage(const PageState_t state_dat, const bool wmode){
  mutex.lock();
  Logsystem::debug_logger()->Tstamp()  << "WStepPage():Locked\n";

  int next_page = (ipage_writing+1) % npage ;
  
  if( status[next_page] == RD_OK && wmode &&
      state_dat != DATA_END ){ // When data is not read yet AND next state != DATAEND.

    Logsystem::debug_logger()->Tstamp()
      << "E DataPage::WStepPage(): next(" << next_page 
      << "pg)Data is not read, Unlocked.\n";
    mutex.unlock();
    usleep(500);
    return -1;
  }
  
  if( status[ipage_writing] == DATA_END){ // Externally Forced to DATA END
    mutex.unlock();
    Logsystem::debug_logger()->Tstamp()
      << "I DataPage::WStepPage(): Found DataEnd Extenal set.\n";
    return -1;
  }

  if( ipage_writing == -1 ){ // In case of the first writing 
    status[0] = state_dat;
  }else{
    Logsystem::debug_logger()->Tstamp()
      << "WStepPage():(" << ipage_writing << "pg)" 
      << PgState2Str(status[ipage_writing]) << " -> "
      << PgState2Str(state_dat) << "\n";
    
    status[ipage_writing] = state_dat;
  }
    

  if(state_dat != DATA_END){
    ipage_writing = next_page;
    Logsystem::debug_logger()->Tstamp() 
      << "WStepPage():(" << ipage_writing << "pg)" 
      << PgState2Str(status[ipage_writing])   << " -> PROTECTED.\n";
    status[ipage_writing] = R_PROTECTED;
  }

  mutex.unlock();
  Logsystem::debug_logger()->Tstamp() << "WStepPage():Unlocked.\n";
  return 0;
}


vector <uint8_t>& DataPage::PagePoint(const int i){
  //  if( i >= (int)dat.size() ){
  if( i >= (int)npage ){
    cerr << "Error in DataPage::PagePoint(): ipage(" << i << ") > npage (" 
	 << npage << ")\n";
    //    return 0;
  }
//   if(npage != dat.size()){
//     cerr << "Error in DataPage::PagePoint(): npage(" << npage 
// 	 << ")  != actual page number(" << dat.size() << ")\n";
//     cerr << "      PageResize() must be done\n";
//     return 0;
//   }
  
  return dat[i];
}

void DataPage::PrintPageState(){
  cerr << "PageState: ";
  for(int i=0; i<npage; i++){
    switch(status[i]){
    case RD_OK:       cerr << "RD_OK "; break;
    case WR_OK:       cerr << "WR_OK "; break;
    case R_PROTECTED: cerr << "R_PRT "; break;
    case EMPTY:       cerr << "EMPTY "; break;
    case DATA_END:    cerr << "DATEND"; break;
    }
  }
  cerr << endl;
}

string DataPage::StrPageState(const int lvl){
  
  string str1 =  "PageState: ";
  for(int i=0; i<npage; i++){
    switch(status[i]){
    case RD_OK:       str1 += "RD_OK  "; break;
    case WR_OK:       str1 += "WR_OK  "; break;
    case R_PROTECTED: str1 += "R_PRT  "; break;
    case EMPTY:       str1 += "EMPTY  "; break;
    case DATA_END:    str1 += "DATEND "; break;
    }
    if(lvl > 0){
      char buf[256]; sprintf(buf,"(%ld)",tarr[i]);
      str1 += buf;
    }
  }
  return str1;
}

void DataPage::InitPages(){   
  mutex.lock();
  fill(status.begin(),status.end(),EMPTY); 
  fill(tarr.begin(),tarr.end(),0);
  ipage_writing = -1;
  mutex.unlock();
}
