
#include "Raw_Proto.hpp"

/** TX1sec -- Trasnmission in single stream, vector version -- */
int32_t RAW_Proto::TX1sec(time_t t, vector <uint8_t>& arr){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }

  //  status = SNDNG;
  int32_t remain = arr.size();
  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  uint8_t* dat_ptr = &arr[0];
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 

  if( remain % pld_size != 0){
    fprintf(stderr,"E @RAW_Proto::TX1sec(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    cerr << "Frame(Payload) size must be ajusted appropriately" << endl;
    return -1;
  }
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize << "\n";
    return -1;
  }

  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);

  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  for(int ipkt=0; ipkt < npkt; ipkt++){
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    dat_ptr += pld_size;   
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
  }
  
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain  << " bytes \n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() 
    << "RAW_Proto::TX1sec(): " << total_sent  << " Bytes sent, end.\n";

  return total_sent;
}


/** TX1secSub -- Trasnmission of sub-channels of data, vector version -- */
int32_t 
RAW_Proto::TX1secSub(const time_t t1, vector <uint8_t>& arr, const DAS_Info* indas){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }

  //  int32_t remain = ExtSubCh(arr,das); // Extracting data for subchannel
  int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
  int32_t remain = arr.size()/ratio;

  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  //  uint8_t* dat_ptr = &preSendBuf[0];
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 

  if( remain % pld_size != 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): bytesec("<< remain
      << ") mod pld(" << pld_size << ") !=0\n"
      << "Frame(Payload) size must be ajusted appropriately.\n";
    return -1;
  }

  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize << "\n";
    return -1;
  }
  vector <uint8_t> databuf(pld_size);
  uint8_t* dat_ptr = &databuf[0];
  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  //  ExtSubCh extch(&arr[0], arr.size(), indas, vlbi_frame->DAS_Inf());
  extch.Setup(&arr[0], arr.size(), indas, vlbi_frame->DAS_Inf());
  bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1;
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  int32_t nbyte(0);
  for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
    if(Is1bitext)  nbyte = extch.Extract1bit(databuf, pld_size);
      else         nbyte = extch.Extract(databuf, pld_size);
    if( nbyte < 0 ){
      cerr << "E @RAW_Proto::Tx1secSub(): Failed extch.Extract()"
	   << " err_cnt=" << err_cnt << endl;
      err_cnt--; 
      if(nbyte < -1) err_cnt = -1;
      continue;
    }
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, nbyte);
    dat_ptr += pld_size;   
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
  }
  
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain 
      << " bytes.\n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): " << total_sent
				      << " Bytes sent, end.\n";

  return total_sent;
}


/** TX1sec -- Trasnmission in single stream, pointer version -- */
int32_t RAW_Proto::TX1sec(time_t t, const uint8_t* arr, const int32_t ndat){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }

  //  status = SNDNG;
  int32_t remain = ndat;
  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  const uint8_t* dat_ptr = arr;
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 

  if( remain % pld_size != 0){
    fprintf(stderr,"E @RAW_Proto::TX1sec(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    cerr << "Frame(Payload) size must be ajusted appropriately" << endl;
    return -1;
  }
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize  << "\n";
    return -1;
  }

  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);

  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  for(int ipkt=0; ipkt < npkt; ipkt++){
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    dat_ptr += pld_size;   
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
  }
  
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain   << " bytes.\n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() 
    << "RAW_Proto::TX1sec(): " << total_sent << " Bytes sent, end.\n";

  return total_sent;
}


/** TX1secSub -- Trasnmission of sub-channels of data, vector version -- */
int32_t 
RAW_Proto::TX1secSub(const time_t t1, const uint8_t* arr, 
		     const int32_t ndat, const DAS_Info* indas){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }

  
  //  status = SNDNG;
  //  int32_t remain = ExtSubCh(arr, ndat, das); // Extracting data for subchannel
  int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
  int32_t remain = ndat/ratio;

  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  //  uint8_t* dat_ptr = &preSendBuf[0];
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 

  if( remain % pld_size != 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): bytesec("<< remain
      << ") mod pld(" << pld_size << ") !=0\n"
      << "Frame(Payload) size must be ajusted appropriately.\n";
    return -1;
  }

  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize << "\n";
    return -1;
  }

  vector <uint8_t> databuf(pld_size);
  uint8_t* dat_ptr = &databuf[0];
  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  //  ExtSubCh extch(arr, ndat, indas, vlbi_frame->DAS_Inf());
  extch.Setup(arr, ndat, indas, vlbi_frame->DAS_Inf());
  bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1;
 
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  int32_t nbyte(0);
  for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
    if(Is1bitext)  nbyte = extch.Extract1bit(databuf, pld_size);
    else           nbyte = extch.Extract(databuf, pld_size);
    
    if( nbyte < 0 ){
      cerr << "E @RAW_Proto::Tx1secSub(): Failed extch.Extract()"
	   << " err_cnt=" << err_cnt << endl;
      err_cnt--; 
      if(nbyte < -1) err_cnt = -1;
      continue;
    }
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, nbyte);
    dat_ptr += pld_size;   
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
      }
      if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
  }
  
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain 
      << " bytes.\n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): " << total_sent
				      << " Bytes sent, end.\n";

  return total_sent;
}


int32_t RAW_Proto::RX_UDP(void* buf, const int32_t buf_size){
  
  int mtu = _sock->get_mtu();
  const int n=3;
  //  status = RCVNG;
  packet.resize(mtu*n);
  memset(&packet[0],0,sizeof(packet));
  memset(buf,0,buf_size);

  const int err_lmt = 10;
  int err_cnt = err_lmt;
  int framesize = vlbi_frame->Framesize();
  int32_t rcv_cnt(0);
  //  , intvl_cnt(0);
  int64_t total_read(0);
  int addr_len = sizeof(addr);
  uint8_t *b0 = (uint8_t*)buf;
  //  int32_t nseq = *pseq;
  //  int32_t npktsps = (int32_t)((double)speed*4/mtu);
  

  while( err_cnt>0 && total_read+framesize <= buf_size ){
    int nread = _sock->recvfrom((struct sockaddr&)addr, addr_len, 
			    (char*)&packet[0], mtu*n);
    if(nread == 0 ) continue;
    
    if(nread != framesize){
      fprintf(stderr, "E @recvfrom(): recevd(%d byte)/expected(%d byte)\n",
	      nread,framesize);
      err_cnt--;
      continue;
    }

    memcpy(b0, &packet[0], nread); 
    total_read += nread;
    b0 += nread;

    err_cnt=err_lmt;
    rcv_cnt++; // intvl_cnt++;

  }// while loop

  Logsystem::error_logger()->Tstamp() 
    << "I @Raw_Proto::RX(): wrote (" << total_read << " Byte).\n";
  return total_read;

}


int RAW_Proto::RXvdif(DataPage& dpage){
    cerr  << "W @Raw_Proto::RXvdif(): is not implementted yet "<< endl;
    return -1;
}

