
/** SUDP Protocol data steam receiver.
 *  $Id: sudp_rxwrite.cpp,v 1.2 2010/01/02 03:57:36 sekido Exp $
 */

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include "SUDPProto.hpp"
#include "Dumper.hpp"

extern "C"{
#include <signal.h>
}

int Dumper::dth_number=0;
int Rate_Meas::thread_number=0;
char *me;  // My name

void usage(){
  fprintf(stderr, "Usage: %s [-p <port>] [-o <file>] [-l <sec>]\n", me);
  fprintf(stderr, "  ----------------------------------------------------\n");
  fprintf(stderr, " -p <port>, default=9901\n");
  fprintf(stderr, " -o <file> Output to file\n");
  fprintf(stderr, " -l <sec> length of receiving data(default=0 means endless).\n");
  fprintf(stderr, " *********** $id$ *********\n");
  exit(1);
}

SUDPProto sudp;

void sighandler(int val){
  sudp.Protocol_stop();
  cerr << "Signal Handler called" << endl;
}

int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 

  int port(DEFAULT_SUDP_PORT);
  string ip("0.0.0.0");
  string fname;
  vector <string> host_list(0);
  int mtu(DEFAULT_MTU_SIZE);  
  uint32_t length(0);
  string logfile;
  
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hp:o:l:")) { 
    case 'p':   port  = atoi(optarg); break;
    case 'l':   length  = atol(optarg); break; // sec
    case 'o':   fname  = optarg; break;
    case 'h':    usage(); break;
    case '?':    usage(); break;
    }
  }

  if(fname.empty()){
    cerr << "Output file is not specified\n"; 
    usage();
  }
  
  string logdir = "/tmp/";
  Logsystem::debug_logger()->enable_logging();
  Logsystem::error_logger()->enable_logging();
  string debug_file = logdir + me + "_dbg.txt";
  Logsystem::debug_logger()->set_stream(debug_file);

  if(! logfile.empty()){
    cerr << "Logging file for debug = " << logfile << endl;
    Logsystem::debug_logger()->enable_logging();
    Logsystem::debug_logger()->set_stream(logfile);
  }

  Logsystem::debug_logger()->Tstamp() << "Logging started \n";

  cerr << "Waiting for packet at port=" << port << endl;

  //  SUDPProto sudp(mtu);
  sudp.Set_mtu(mtu);
  sudp.Set_TransLength(length); // sec
  sudp.Timer_Start();
  sudp.Disp_RxRate();
  VDIF_Frame vdif_frame;
  sudp.Set_VlbiFrame(&vdif_frame);
  if(! sudp.Rx_Ready(port)){
    cerr << " Failed sudp.Rx_Ready("<< port <<")" << endl;
    exit(1);
  }

  time_t t_start, t_now;
  time(&t_start);
  t_now = t_start;
  
  FILE *fp(0);
  cout << "Receiving to File (" << fname << ")\n";

  if((fp= fopen(fname.c_str(), "w")) ==NULL){
    cerr << "Error: Filed to open file(" << fname << ") \n";
    exit(1);
  }
  signal(SIGUSR1,&sighandler);
  
  int32_t nwrote = sudp.RX_SUDP(fp);
  fprintf(stderr," Wrote to file(%s) %ld bytes\n", 
	  fname.c_str(), (long int)nwrote);
  sudp.Finish();
  fclose(fp);
  
  signal(SIGUSR1,SIG_DFL);

  //  sudp.Timer_Stop(); // Rate measurement stop

  Logsystem::debug_logger()->Tstamp() << "Logging end.\n";
  //  Logsystem::debug_logger()->close_stream();
  //  delete Logsystem::debug_logger();

  cerr << "Recerving End.\n";
}



