
/** @file vdif_chk is test reading VDIF header data 
 *  @author M.Sekido
 *  @Date 2009.4.6
 */

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>

#include "VDIF_Header.hpp"
#include "DAS_info.hpp"
#include "VsiData.hpp"
#include "hist_g.hpp"

extern "C"{
#include <unistd.h>
}

char *me;

void usage(){
  cerr << "Usage: " << me << " -f <filename> [-l <level>]\n";
  //  cerr << "  -d <n>: Deltailed output level (default: 0)\n";
  cerr << "-l : Display level 0-4\n";
  exit(1);
}

int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */

  string fname;
  //  if(argc < 2) usage();
  int dlevel=1;

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:l:")) { 
    case 'h':  usage(); break;
    case 'f':    fname = optarg;      break;
    case 'l':    dlevel = atoi(optarg);      break;
      //    case 'd': vdif_hdr.Set_dbg(atoi(optarg)); break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(fname.empty()) usage();
  
   FILE *fp = fopen(fname.c_str(),"r");
  if(fp ==NULL){
    cerr << "Error in opening file: " << fname << endl;
    exit(1);
  }
  
  VDIF_Hdr vdifhdr[2];
  VDIF_Hdr vdfh;
  int headsize = vdfh.Headsize();
  fprintf(stderr, "VDIF headsize=%lu %d\n", sizeof(vdfh),
	  headsize);

  int offset(0);
  
  VDIF_Hdr * pvdifhdr =      VDIF_Hdr_find(fp,offset);
  int icnt=0;
  int jcnt=0;

  Time_dat t1; 


  int nch = pvdifhdr->Nch();
  int qbit = pvdifhdr->Qbit();
  int32_t srate = pvdifhdr->Srate();
  DAS_Info das(srate,nch,qbit);
  VsiData vsidat;
  vsidat.Set_DasInfo(&das);
  int imax(0),imin(0), istep(0);
  switch(qbit){
  case 1: imax= 0; imin = -1, istep=1; break;
  case 2: imax= 3; imin = -3, istep=2; break;
  case 4: imax= 8; imin = -8, istep=1; break;
  case 8: imax= 128; imin = -128, istep=1; break;
  }

  Hist_g <int8_t> hist(imin,imax,istep);
  
  while(pvdifhdr !=0){
    vdfh = * pvdifhdr;
    if(pvdifhdr->iFrame() ==0){
      if(jcnt==0){
	vdifhdr[0] = *pvdifhdr; icnt=0; jcnt++;
      }else{
	vdifhdr[1] = *pvdifhdr;
	double dt = vdifhdr[1].Sec_Ref() - vdifhdr[0].Sec_Ref();
	double Mbps = (double)(icnt * vdifhdr[1].Framesize())/dt*8*1.e-6;
	srate = (int32_t)round(Mbps);
      }
    }      
    
    if(t1 != vdfh.Epoch() || dlevel > 2){
      vdfh.Print_hd(stderr);
    }
    if( dlevel > 3) 
      fprintf(stderr,"Hdr offset(%d) ",offset+headsize);
    if(dlevel > 4){
      if(vsidat.Read1Sec(fp)< 0){
	cerr << " Error in Read1Sec()\n";
      };
      vsidat.Make_Table();
      vsidat.Interpret();
      for(int ich =0; ich < nch; ich++){
	cerr << "\n Hist("<< ich << "): ";
	hist.Init();
	hist.Accm(vsidat.ValArr(ich));
	vector <float>& arr = hist.Ret_RatArr();
	for(int i=0; i< (int)arr.size(); i++) cerr << arr[i] <<" ";
      }
      cerr << endl;
    }else{
      fseek(fp,headsize,SEEK_CUR);      
    }
    if( dlevel > 5){
      if(srate >0) fprintf(stderr," Est rate= %d Mbps, ", srate);
    }

    t1 = vdfh.Epoch();    

    pvdifhdr = VDIF_Hdr_find(fp,offset);

    icnt++;
  }
  
  fclose(fp);
  return 0;
}

