/**
 * @file  tds_cls.cpp
 * @brief TDS sampler (K5/VSSP) data handling abstract class
 *
 * @author M.Sekido (NICT Kashima.)
 * @date $Date: 2010/01/02 03:40:37 $
 * @version $Id: tds_cls.cpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 * $Log: tds_cls.cpp,v $
 * Revision 1.2  2010/01/02 03:40:37  sekido
 * Replace FirstBranch to main
 *
 * Revision 1.1.2.3  2009/10/06 10:12:47  vlbi
 * debug sudp_send to run.
 *
 * Revision 1.1.2.2  2009/09/28 10:32:07  sekido
 * sudp_recv2 compile.
 *
 * Revision 1.1.2.1  2009/09/09 04:31:15  vlbi
 * sudp_send created and compiled.
 *
 * Revision 1.2  2009/08/27 01:42:31  sekido
 * Implemented statics_vssp32.cpp and Debugged 8bit encoding in tds_cls.cpp.
 *
 * Revision 1.1  2009/05/28 12:19:11  sekido
 * Registering appropriate selected source files.
 *
 * Revision 1.1  2007/05/14 06:37:03  sekido
 * tds_cls.cpp $B$NDI2C$rK:$l$F$$$?!#(B CVS$B$G4IM}$r3+;O$9$k!#(B
 *
 * 
 * Copyright (C) NICT Kashima. All rights reserved.
 */

#include "assert.h"

#include "tds_cls.hpp"
#include "time_lib.hpp"
//using namespace tds;



const int32_t TDS_DAT::Interprit_1bit(const uint32_t ndat_request){
  
  uint32_t  cnt=0;
  uint32_t nofst;
  float* pv=0;
  vector <unsigned char>::iterator & buf_pointer = Buffer_pointer();
  uint32_t progres_point = ndat_request/10;
  while(progres_point % 8 != 0) progres_point --;

  while( cnt < ndat_request ){
    /*buffer always have to be used by exact number */
    assert( buf_pointer < Buffer_End()); 
    
    if( nch ==1){
      int ich =0;
      for(int j=0; j<4 && buf_pointer <  Buffer_End(); j++){
	nofst = (*buf_pointer) * 8;
	if( Farr_Set(ich, 8, pValue_tbl(nofst)) <0 ){
	  fprintf(stderr, 
		  "Error:End of farr[%d] in Interprit_1bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);  exit(1); }	  
	buf_pointer++;
	cnt+=8;
      } // for loop 

      //      fremain +=32;
    }else{
      if( nch != 4) {
	fprintf(stderr, 
		"Error: nch(%d) != 4. This software assume nch must be 4\n", 
		nch);
	exit(EXIT_FAILURE);
      }
      //      float *pf[4] = {pFarr_tail(0), pFarr_tail(1),pFarr_tail(2),pFarr_tail(3)}; 
      for(int j=0; j<4; j++){
	nofst = (unsigned int)(*buf_pointer) * 8; 
	pv = pValue_tbl(nofst);
	int ich=0;
	for(int i=0; i < 8; i++){
	  ich = ich % 4;
	  if( Farr_Set(ich, 1, pv) <0){
	    fprintf(stderr, "Error:End of farr[%d] in Interprit_1bit() at cnt=%d/%d.\n", 
		    ich,cnt,ndat_request);	  exit(1); }	  
	  pv++; ich++;
	  //  	  *pf[ich] = *(pv+ich);     pf[ich]++;
	  //  	  *pf[ich] = *(pv + ich+nch); pf[ich]++;
	}
	buf_pointer++;
	cnt +=2;
      }
      //        for(int ich=0; ich < nch; ich++)	
      //  	if(!pFarr_inc(ich,8)){
      //  	  fprintf(stderr, "Errpr:pFarr_inc() failed at cnd=%d/%d\n", 
      //  		  cnt,ndat_request);
      //  	  exit(1);
      //  	}
      //      fremain +=8;
    }
    if(cnt % progres_point == 0) fputc('.',stderr);
    if(cnt == ndat_request ) break;
    if( (ndat_request - cnt) % (32/nch/sbit)  != 0 ){
      fprintf(stderr,"Error: Reuested data number (%d) - counted (%d) is not multiple of %d bit (nch=%d),\n", 
	      ndat_request,cnt, 32/nch/sbit, nch);
      cerr << " something wrong. Probably time span or data number is not multiple of 32bit data\n";
      exit(EXIT_FAILURE);
    }
  }

  double dt = 1./(double)Sfrq();
  vector <double>::iterator& pdxe = pTarr_tail();
  double tmp_time = Time_elapsed();
  for(uint i=0; i<cnt; i++, pdxe++ ){
    *pdxe = tmp_time;
    tmp_time += dt; 
  }
  /* Time(sec) counted from t0*/
  double _time_adv = (double)cnt/Sfrq();
  Add_time_elapsed(_time_adv);
  
  return(cnt);

}

const int32_t TDS_DAT::Interprit_2bit(const uint32_t ndat_request){

  uint32_t  nofst;
  float* pv;
  vector <unsigned char>::iterator & buf_pointer = Buffer_pointer();

  uint32_t cnt=0;
  while( cnt < ndat_request){
    /*buffer always have to be used by exact number */
    assert( buf_pointer < Buffer_End()); 
    if( nch==1){
      int ich =0;
      for(int j=0; j<4 && buf_pointer < Buffer_End(); j++){
	nofst = (*buf_pointer) * 4;
	pv = pValue_tbl(nofst);
	if( Farr_Set(ich, 4, pv) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_2bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);	  exit(1); }
	//	memcpy( (void *)pFarr_tail(0), pValue_tbl(nofst), sizeof(float)*4);
//  	if( !pFarr_inc(0,+4)){
//  	  fprintf(stderr, 
//  		  "Error:End of farr[%d] in Interprit_2bit() at cnt=%d/%d.\n", 
//  		  0, cnt,ndat_request);
//  	  exit(1);
//  	}
	buf_pointer++;
      }
      cnt+=16;
      //      fremain +=16;
    }else{
      if( nch != 4) {
	fprintf(stderr, 
		"Error: nch(%d) != 4. This software assume nch must be 4\n", 
		nch);
	exit(EXIT_FAILURE);
      }
      
      //      float *pf[4] = {pFarr_tail(0),pFarr_tail(1), pFarr_tail(2),pFarr_tail(3)}; 
      for(int j= 0;j<4; j++){
	nofst = (*buf_pointer)*4;
	pv = pValue_tbl(nofst); 
	for(int ich= 0;ich<nch; ich++){
	  if( Farr_Set(ich, 1, pv) <0){
	    fprintf(stderr, "Error:End of farr[%d] in Interprit_2bit() at cnt=%d/%d.\n", 
		    ich,cnt,ndat_request);	  exit(1); }
	  //	  *pf[ich] = *pv;
	  pv++;   // pf[ich]++;
	}
	buf_pointer++;
      }
      //        for(int ich= 0;ich<nch; ich++)
      //  	if( !pFarr_inc(ich,4)){
      //  	  fprintf(stderr, "Errpr:pFarr_inc() failed at cnd=%d/%d\n",
      //  		  cnt, ndat_request);
      //  	  exit(1);
      //  	}
      cnt +=4;
      //      fremain +=4;
    } /* 1ch or 4ch */
    
    if( cnt == ndat_request ) break;
    if( (ndat_request - cnt) % (32/nch/sbit)  != 0 ){
      fprintf(stderr,"Error: Reuested data number (%d) - counted (%d) is not multiple of %d bit (nch=%d),.\n", 
	      ndat_request,cnt, 32/nch/sbit, nch);
      cerr << "something wrong. Probably time span or data number is not multiple of 32bit data\n";
      exit(EXIT_FAILURE);
    }
  } /* While loop */
  
  double dt = 1./(double)Sfrq();
  vector <double>::iterator& pdxe = pTarr_tail();
  double tmp_time = Time_elapsed();
  for(uint i=0; i<cnt; i++, pdxe++ ){
    *pdxe = tmp_time;
    tmp_time += dt; 
  }
  //  for(int i=0; i<cnt; i++, pdxe++)  *pdxe = Time_elapsed() + (double)i*dt ; 
  /* Time(sec) counted from sec0*/
  double _time_adv = (double)cnt/Sfrq();
  Add_time_elapsed(_time_adv);

  return(cnt);
}

const int32_t TDS_DAT::Interprit_4bit(const uint32_t ndat_request){


  float* pv;
  unsigned nofst;
  vector <unsigned char>::iterator & buf_pointer = Buffer_pointer();

  uint32_t cnt=0;
  while( cnt < ndat_request){
    /*buffer always have to be used by exact number */
    assert( buf_pointer < Buffer_End()); 
    
    if(nch==1){
      int ich =0;
      for(int j=0; j<4 && buf_pointer < Buffer_End(); j++){
	nofst = (*buf_pointer)*2;
	if( Farr_Set(ich, 2, pValue_tbl(nofst)) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_4bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);	  exit(1); }
	
	//	memcpy( (void *)pFarr_tail(0), pValue_tbl(nofst), sizeof(float)*2);
	//  	if( !pFarr_inc(0,+2)){
	//  	  fprintf(stderr, 
	//  		  "Error:End of farr[%d] in Interprit_4bit() at cnd=%d/%d.\n", 
	//  		  0,cnt,ndat_request);
	//  	  exit(1);
	//  	}
	buf_pointer++;
      }
      cnt+=8;
      //      fremain +=8;
    }else{ 
      if( nch != 4) {
	fprintf(stderr, 
		"Error: nch(%d) != 4. This software assume nch must be 4\n", 
		nch);
	exit(EXIT_FAILURE);
      }
      //      float *pf[4] = {pFarr_tail(0), pFarr_tail(1),pFarr_tail(2),pFarr_tail(3)}; 
      int ich=0;
      for(int j=0; j<4; j++){
	nofst = (*buf_pointer)*2;
	pv = pValue_tbl(nofst); 
	ich = ich %4;
	if( Farr_Set(ich, 1, pv) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_4bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);	  exit(1); }
	ich++; pv++;
	if( Farr_Set(ich, 1, pv) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_4bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);  exit(1);	}
	ich++;
	buf_pointer++;
      }
//        for(int ich= 0;ich<nch; ich++) 
//  	if(!pFarr_inc(ich,2)){
//  	  fprintf(stderr, 
//  		  "Error:End of farr[%d] in Interprit_4bit() at cnt=%d/%d.\n", 
//  		  0,cnt,ndat_request);
//  	  exit(1);	
//  	}
      cnt+=2;
      //      fremain +=2;
      if( cnt == ndat_request) break;
    } // case of 4ch  

    if( (ndat_request - cnt) % (32/nch/sbit)  != 0 ){
      fprintf(stderr,"Error: Reuested data number (%d) - counted (%d) is not multiple of %d bit (nch=%d).\n", 
	      ndat_request, cnt, 32/nch/sbit,nch);
      cerr <<", something wrong. Probably time span or data number is not multiple of 32bit data\n";
      exit(EXIT_FAILURE);
    }
  }// while loop
  
  double dt = 1./(double)Sfrq();
  vector <double>::iterator& pdxe = pTarr_tail();
  //  for(int i=0; i<cnt; i++, pdxe++)  *pdxe =Time_elapsed() +  (double)i*dt; 
  double tmp_time = Time_elapsed();
  for(uint i=0; i<cnt; i++, pdxe++ ){
    *pdxe = tmp_time;
    tmp_time += dt; 
  }
  /* Time(sec) counted from sec0*/
  double _time_adv = (double)cnt/Sfrq();
  Add_time_elapsed(_time_adv);

  return(cnt);
}

const int32_t TDS_DAT::Interprit_8bit(const uint32_t ndat_request){

  vector <unsigned char>::iterator &buf_pointer = Buffer_pointer();
  
  uint32_t cnt=0;
  while( cnt < ndat_request){
    /*buffer always have to be used by exact number */
    assert( buf_pointer < Buffer_End()); 

    if(nch==1){
      int ich=0;
      for(int j=0; j<4 && buf_pointer < Buffer_End(); j++){
	if( Farr_Set(ich, 1, pValue_tbl(*buf_pointer)) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_8bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);	  exit(1); }
	//  	if( !pFarr_inc(0,+1)){
	//  	  fprintf(stderr,
	//  		  "Error:End of farr[%d] in Interprit_8bit() at cnt=%d/%d.\n", 
	//  		  0,cnt,ndat_request);
	//  	  exit(1);
	//  	}
	buf_pointer++;
      }
      cnt+=4;
      //      fremain +=4;
    }else{ 
      if( nch != 4) {
	fprintf(stderr, 
		"Error: nch(%d) != 4. This software assume nch must be 4\n",
		nch);
	exit(EXIT_FAILURE);
      }
      for(int ich=0; ich<4; ich++){
	if( Farr_Set(ich, 1, pValue_tbl(*buf_pointer)) <0){
	  fprintf(stderr, "Error:End of farr[%d] in Interprit_8bit() at cnt=%d/%d.\n", 
		  ich,cnt,ndat_request);	  exit(1); }
//  	*pFarr_tail(ich)  = *pValue_tbl(*buf_pointer);
//  	if(!pFarr_inc(ich,1)){
//  	  fprintf(stderr,
//  		  "Error:End of farr[%d] in Interprit_8bit() at cnt=%d/%d.\n", 
//  		  0,cnt,ndat_request);
//  	  exit(1);
//  	}
	buf_pointer++;
      }
      cnt++;
      //      fremain ++;
    }
    if( cnt == ndat_request) break;
    if( (ndat_request - cnt) % (32/nch/sbit)  != 0 ){
      fprintf(stderr,"Error: Reuested data number (%d) - counted (%d) is not multiple of %d bit (nch=%d).\n", 
	      ndat_request,cnt, 32/nch/sbit,nch);
      cerr << ", something wrong. Probably time span or data number is not multiple of 32bit data\n";
      exit(EXIT_FAILURE);
    }
  }

  double dt = 1./(double)Sfrq();
  vector <double>::iterator& pdxe = pTarr_tail();
  //  for(int i=0; i<cnt; i++,pdxe++)  *pdxe =Time_elapsed() + (double)i*dt ; 
  double tmp_time = Time_elapsed();
  for(uint i=0; i<cnt; i++, pdxe++ ){
    *pdxe = tmp_time;
    tmp_time += dt; 
  }
  /* Time(sec) counted from sec0*/

  double _time_adv = (double)cnt/Sfrq();
  Add_time_elapsed(_time_adv);

  return(cnt);

}


void TDS_DAT::Set_Valtbl(void){
  
  int narr(0);
  //    unsigned int msk[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};
  
  switch (sbit) {
  case 1: narr = 256*8; break;
  case 2: narr = 256*4; break;
  case 4: narr = 256*2; break;
  case 8: narr = 256; break;
  default:
    fprintf(stderr, "Error in TDS_DAT::Set_Valtbl(): sbit(%d) is not 1,2,4,8\n",
	    sbit);
  }
  value_tbl.resize(narr);
  vector <float>::iterator pf = value_tbl.begin();
    
  switch (sbit){
  case 1: 
    for(int i=0; i<256; i++){
      for(int j=0; j<8; j++){
	if( (i >> j ) & MASK_1BIT) *pf = 1.; else  *pf = -1.; 
	pf++;
      }
    }
    break;
  case 2:
    for(int i=0; i<256; i++){
      for(int j=0; j<8; j+=2){
	switch ( (i >> j) & MASK_2BIT){
	case 0:   *pf = -3.; break;
	case 1:   *pf = -1.; break;
	case 2:   *pf =  1.; break;
	case 3:   *pf =  3.; break;
	}  
	//	  fprintf(stderr, "*pf =%f ", *pf);
	pf++;
      }
      //	putc('\n', stderr);
      // getchar();
    }
    break;
  case 4:
    for(int i=0; i<256; i++){
      for(int j=0; j<8; j+=4){
	*pf = (float)( (i >> j) & MASK_4BIT) -8.;
	pf++;
      }
    }
    break;
  case 8:
    for(int i=0; i<256; i++){
      *pf = (float)i -128.; pf++;
    }
    break;
  }
    

}

void TDS_HD::Print_inf(FILE *fp, const int level){
//   if(data[3] == 0x8b || data[3] == 0x8c )
//     fwrite("HD_Symbol OK\n",13,1,fp);
//   else
  switch(data[3]){
  case 0x8b: 
    //    fwrite("   HD_Type=VSSP\n",16,1,fp);
    fprintf(fp," HD_Type=VSSP");
    break;
  case 0x8c:
    //    fwrite("   HD_Type=VSSP32\n",18,1,fp);
    fprintf(fp," HD_Type=VSSP32");
    break;
  default:
    //    fwrite("   HD_Symbol NG\n",16,1,fp);
    fprintf(fp," HD_Symbol NG");
    break;
  }
  if( level > 0 ){
    int yy = Year();
    int doy = Doy();
    int sod = Sod();
    Time_dat t1(yy,doy,0,0,0);
    t1 += sod;
    fprintf(fp, " Epoch: %s ",t1.Print_time(10).c_str());
    fprintf(fp,"  %2d Msps  %d bit %2d ch ",
	    Sfrq()/1000000, Bit(), Nch());
    fprintf(fp,"Fmt:%d, Station: %s(%s) ",
	    FormatNum(), Stname().c_str(),StnID().c_str() );
    fprintf(fp,"Hostname: %s", Hostname().c_str() );
  }
  putc('\n',fp);
}


/*
 * $Log: tds_samplr.hpp
 *  Author: $Author: sekido $
 *  Date: $Date: 2010/01/02 03:40:37 $
 *  $Header: /vsspfs1/work3/cvs-archive/VDIF/tds_cls.cpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 *  Revision: $Revision: 1.2 $
 *  Source: $Source: /vsspfs1/work3/cvs-archive/VDIF/tds_cls.cpp,v $
 */
